# GTK+ and related
import gtk

# Python modules
import re, os, os.path, shelve, tempfile

# Gnuplot-related modules
import gnuplot, plot
from gnuplot import Gnuplot

# Interface modules
import window, export, render, preferences, appearances, conformal, colorspec, axes, multiplot

# Helper modules
import config
from session import Session

class Main(window.Window):
	def __init__(self):
		window.Window.__init__(self, "main")
		
		s = Session()
		s.plot_hierarchy = self.plot_hierarchy
		s.status = self.status
		
		ts_plot = self.child("ts_plot")
		ts_plot.connect("row-inserted", self.on_ts_plot_changed)
		ts_plot.connect("row-deleted", self.on_ts_plot_changed)
		ts_plot.connect("row-changed", self.on_ts_plot_changed)
		ts_plot.connect("rows-reordered", self.on_ts_plot_changed)

		tsel = self.child("tv_plot").get_selection()
		tsel.set_mode(gtk.SELECTION_MULTIPLE)
		tsel.connect("changed", self.on_tsel_plot_changed)

		self.set_embed_view()

	def add_plot(self, plot, parent = None):
		if plot:
			attr = "%s%s" % ("surface" if plot.is_surface else "2d curve", ", parametric" if plot.is_parametric else "")
			title = "%s <i>(%s)</i>"%(plot.title, attr)

			tm = self.child("ts_plot")
			last = tm.append(parent, (title, self.assign_plot_uid(plot), True))

			tv = self.child("tv_plot")
			if parent:
				tv.expand_row(tm.get_path(parent), False)
			tsel = tv.get_selection()
			tsel.unselect_all()
			tsel.select_iter(last)
			tv.scroll_to_cell(tm.get_path(last))

			self.plot_list()

	def set_embed_view(self):
		gp = Gnuplot()
		gp.set_terminal("x11", 'window "%x"' % self.builder.get_object("dwg_gnuplot").window.xid)

	def get_cwd(self):
		gp = Gnuplot()
		return gp.get_last_event().strip() if gp.pwd() != None else os.curdir

	def gui2plot(self):
		func = self.child("txt_function").get_text()
		if not func:
			self.dlg_error("Please type a formula or select a valid data file.")
			return None

		p = plot.Plot()

		entries = ["title", "function", "index", "every", "using", "lc"]
		spins   = ["ls", "lt", "lw", "ps", "pt"]
		combos  = ["style", "fs", "smooth", "matrix"]
		checks  = ["ls", "parametric", "palette", "matrix"]
		radios  = ["surface"]

		for entry in entries:
			exec('p.%s = self.child("txt_%s").get_text().strip()' % (entry, entry))

		for spin in spins:
			exec('p.%s = self.child("spb_%s").get_value()' % (spin, spin))

		for combo in combos:
			exec('p.%s = self.child("cbo_%s").get_active_text().strip()' % (combo, combo))
			exec('p.%s_index = self.child("cbo_%s").get_active()' % (combo, combo))

		for check in checks:
			exec('p.is_%s = self.child("chk_%s").get_active()' % (check, check))

		for radio in radios:
			exec('p.is_%s = self.child("rdo_%s").get_active()' % (radio, radio))

		if re.match(r"""["'](.+)["']""", p.function) != None:
			filename = p.function
			absfile = "%s/%s" % (self.get_cwd(), filename.strip('"'))
			if not (os.path.exists(filename.strip("'\"")) or os.path.exists(absfile)):
				self.dlg_error("The data file does not exist.")
				return None
			p.is_datafile = True
		else:
			p.is_datafile = False

		if not p.title: p.title = p.function.strip('"') if p.is_datafile else p.function

		return p if p.test(self.window) == None else None

	def plot2gui(self, p):
		entries = ["title", "function", "index", "every", "using", "lc"]
		spins   = ["ls", "lt", "lw", "ps", "pt"]
		combos  = ["style", "fs", "smooth", "matrix"]
		checks  = ["ls", "parametric", "palette", "matrix"]
		#radios  = ["surface"]

		for entry in entries:
			exec('self.child("txt_%s").set_text(p.%s if p.%s else "")' % (entry, entry, entry))

		for spin in spins:
			exec('self.child("spb_%s").set_value(int(p.%s) if p.%s else 0)' % (spin, spin, spin))

		for combo in combos:
			exec('self.child("cbo_%s").set_active(p.%s_index if p.%s else 0)' % (combo, combo, combo))

		for check in checks:
			exec('self.child("chk_%s").set_active(p.is_%s)' % (check, check))

		if p.is_surface: self.child("rdo_surface").set_active(True)
		else: self.child("rdo_curve").set_active(True)

	def assign_plot_uid(self, plot):
		s = Session()
		
		s.plot_dict[s.plot_uid] = plot
		retval = s.plot_uid
		s.plot_uid = s.plot_uid + 1
		return retval

	def plot_list(self):
		ts = self.child("ts_plot")
		it = ts.get_iter_first()
		cmd = ""
		is_multiplot = self.child("mnu_mode_multiplot").get_active()
		is_parametric = self.child("chk_parametric").get_active()
		is_surface = self.child("rdo_surface").get_active()
		gp = Gnuplot()
		s = Session()
		counter = 0

		if not it: return -1

		while it:
			if not ts.get_value(it, 2):
				it = ts.iter_next(it)
				continue
			plot = s.plot_dict[ts.get_value(it, 1)]
			if (plot.is_parametric, plot.is_surface) != (is_parametric, is_surface):
				it = ts.iter_next(it)
				continue
			cmd = "%s, %s" % (cmd, repr(plot)) if cmd != "" else repr(plot)
			counter = counter + 1
			if ts.iter_has_child(it):
				child = ts.iter_children(it)
				while child:
					if not ts.get_value(child, 2):
						child = ts.iter_next(child)
						continue
					plot = s.plot_dict[ts.get_value(child, 1)]
					if (plot.is_parametric, plot.is_surface) != (is_parametric, is_surface):
						child = ts.iter_next(child)
						continue
					cmd = '%s, %s' % (cmd, repr(plot))
					counter = counter + 1
					child = ts.iter_next(child)

			it = ts.iter_next(it)
			
			if is_multiplot and cmd != "": # Multiplot mode
				if s.multiplot_count >= s.multiplot_total:
					print "breaking"
					break
				gp.set_origin(s.multiplot_data[s.multiplot_count][0], s.multiplot_data[s.multiplot_count][1])
				gp.cmd("set size %s, %s" % (s.multiplot_data[s.multiplot_count][2], s.multiplot_data[s.multiplot_count][3]))
				s.multiplot_count = s.multiplot_count + 1
				gp.clear()
				gp.plot(cmd) if not is_surface else gp.splot(cmd)
				if gp.notify_last(self.window): return 1
				cmd = ""

		if not is_multiplot:
			if cmd == "":
				gp.clear()
				self.sensitiveness(["mnu_file_export", "mnu_file_render"], False)
			else:
				gp.plot(cmd) if not is_surface else gp.splot(cmd)
				self.sensitiveness(["mnu_file_export", "mnu_file_render"], True)
				if gp.notify_last(self.window) is not None: return -1

		return counter

	def plot_hierarchy(self):
		ts = self.child("ts_plot")
		it = ts.get_iter_first()

		if not it: return None

		ph = []
		while it:
			ch = []
			ch.append(tuple(ts.get(it, 0, 1, 2)))
			if ts.iter_has_child(it):
				child = ts.iter_children(it)
				while child:
					ch.append(tuple(ts.get(child, 0, 1, 2)))
					child = ts.iter_next(child)
			ph.append(ch)
			it = ts.iter_next(it)
		return ph

	def plot_from_hierarchy(self, ph):
		ts = self.child("ts_plot")
		for block in ph:
			parent = ts.append(None, block[0])
			for plot in block[1:]:
				ts.append(parent, plot)

	def plot_from_last_plot_command(self):
		gp = Gnuplot()
		gp.show("plot")
		plot_str = gp.get_last_event()
		is_surface = plot_str.find("was: s") != -1
		gp.show("parametric")
		is_parametric = gp.get_last_event().find("ON") != -1
		plot_str = plot_str[plot_str.find("was: ") + 5:]
		plots = re.match("(.+?) (.*)", plot_str.strip()).group(2).split(',')

		for pl in plots:
			p = plot.Plot()
			p.parse(pl)
			p.is_surface = is_surface
			p.is_parametric = is_parametric
			self.add_plot(p)

	def update_window_title(self):
		s = Session()
		
		try:
			path, fn = os.path.split(s.filename)
		except AttributeError:
			fn = "Unsaved session"
			path = ""
#		try:
#			match = re.match("(.*)[/](.*)", s.filename)
#			fn = match.group(2)
#			path = " (%s)" % match.group(1)
#		except:
#			fn = "Unsaved session"
#			path = ""
		self.window.set_title("%s%s%s - pyGtkPlot" % ("*" if s.has_changed else "", fn, " (%s)" % path if path and path != "" else ""))

	def status(self, text):
		self.child("sb_main").push(0,text)

	def clear(self):
		self.child("ts_plot").clear()

	def populate_history(self, needle):
		s = Session()
		
		ls = self.child("ls_history")
		ls.clear()
		for entry in s.history:
			if entry.find(needle) == 0: last = ls.append([entry])

		return last

	def show_history(self, it):
		rect = txt = self.child("txt_raw_cmd").get_allocation()
		pos = self.window.get_position()

		hist = self.child("history")
		hist.show_all()

		tv = self.child("tv_history")
		tv.get_selection().select_iter(it)
		tv.scroll_to_cell(tv.get_model().get_path(it))
		tv.grab_focus()

		hist.move(pos[0] + rect.x, pos[1] + rect.y + rect.height)
		hist.set_size_request(rect.width, 96)

### Signal handlers ############################################################

	def on_tsel_plot_changed(self, tsel, data = None):
		tm, sel = tsel.get_selected_rows()

		self.sensitiveness(["vb_plot", "cmd_add_child", "cmd_apply", "mnu_edit_add_child", "mnu_edit_apply"], len(sel) <= 1)
		self.sensitiveness(["cmd_remove", "mnu_edit_remove", "mnu_popup_remove"], len(sel) > 0)

	def on_cmd_add_clicked(self, widget, data = None):
		self.add_plot(self.gui2plot())

	def on_cmd_add_child_clicked(self, widget, data = None):
		plot = self.gui2plot()
		if plot:
			tm, paths = self.child("tv_plot").get_selection().get_selected_rows()
			ti = tm.get_iter(paths[0])
			if ti:
				tparent = tm.iter_parent(ti)
				if tparent:
					ti = tparent
			self.add_plot(plot, ti)

	def on_cmd_apply_clicked(self, widget, data = None):
		plot = self.gui2plot()
		if not plot: return
		tm, path = self.child("tv_plot").get_selection().get_selected_rows()
		it = tm.get_iter(path[0])
		attr = "%s%s" % ("surface" if plot.is_surface else "2d curve", ", parametric" if plot.is_parametric else "")
		title = "%s <i>(%s)</i>"%(plot.title, attr)
		tm.set_value(it, 0, title)
		s = Session()
		s.plot_dict[tm.get_value(it, 1)] = plot
		self.plot_list()

	def on_cmd_color_clicked(self, widget, data = None):
		lc = colorspec.Colorspec()
		lc.set_title("Line color specification")
		ans = lc.run()
		if ans == gtk.RESPONSE_OK:
			self.child("txt_lc").set_text(lc.get_color())
		lc.destroy()
		
	def on_cmd_remove_clicked(self, widget, data = None):
		s = Session()
		
		ts, paths = self.child("tv_plot").get_selection().get_selected_rows()
		paths.reverse()
		for path in paths:
			it = ts.get_iter(path)
			if it:
				if ts.iter_has_child(it):
					child = ts.iter_children(it)
					while child:
						del s.plot_dict[ts.get_value(child, 1)]
						child = ts.iter_next(child)
				del s.plot_dict[ts.get_value(it, 1)]
				ts.remove(it)
		self.plot_list()

	def on_cmd_replot_clicked(self, widget, data = None):
		self.plot_list()

	def on_cmd_send_clicked(self, widget, data = None):
		gp = Gnuplot()
		conf = config.Config()
		s = Session()
		
		cmd = self.child("txt_raw_cmd")
		text = cmd.get_text()
		gp.notify(self.window, text)
		conf = config.Config()
		max_line = int(float(conf.max_history))
		if max_line:
			if s.history == [] or s.history[-1] != text:
				s.history.append(text)
				if len(s.history) > max_line > 0:
					s.history = s.history[-max_line:]
		else:
			s.history = []
		cmd.set_text("")
		s.has_changed = True
		self.update_window_title()
		
		gp.show("plot")
		if gp.last_info.find("was:\n") == -1:
			if (text[0:1] == "p" or text[0:2] == "sp") and conf.add_from_cmd_line == "True":
				self.plot_from_last_plot_command() 
			elif conf.replot_at_cmd == "True":
				gp.replot()

	def on_ts_plot_changed(self, treemodel, it, data = None):
		if not treemodel.iter_n_children(None):
			gp = Gnuplot()
			gp.clear()
			self.sensitiveness(["cmd_remove", "cmd_replot", "cmd_apply", "cmd_add_child",\
			                    "mnu_edit_remove", "mnu_edit_apply", "mnu_edit_add_child",\
			                    "mnu_edit_select_all", "mnu_edit_clear",\
			                    "mnu_popup_remove", "mnu_popup_select_all", "mnu_popup_clear"], False)
		else:
			self.sensitiveness(["cmd_replot", "mnu_edit_select_all", "mnu_edit_clear",\
			                    "mnu_popup_select_all", "mnu_popup_clear"], True)

		s = Session()
		s.has_changed = True
		self.update_window_title()

	def on_tv_plot_cursor_changed(self, widget, data = None):
		tm, it = widget.get_selection().get_selected_rows()
		if len(it):
			s = Session()
			self.sensitiveness(["cmd_apply", "cmd_add_child",\
			                    "mnu_edit_apply", "mnu_edit_add_child",], True)
			self.plot2gui(s.plot_dict[tm.get_value(tm.get_iter(it[0]), 1)])
		else:
			self.sensitiveness(["cmd_remove", "cmd_apply", "cmd_add_child",\
			                    "mnu_edit_remove", "mnu_edit_apply", "mnu_edit_add_child",\
			                    "mnu_popup_remove"], False)

	def on_tv_plot_button_press_event(self, widget, event, data = None):
		if (event.type, event.button) == (gtk.gdk.BUTTON_PRESS, 3):
			self.child("mnu_popup").popup(None, None, None, event.button, event.time)
			return True

	def on_cr_visible_toggled(self, widget, path, data = None):
		tm = self.child("ts_plot")
		it = self.child("ts_plot").get_iter(path)

		tm.set_value(it, 2, not tm.get_value(it, 2))

		self.plot_list()

	def on_delete_event(self, widget, data = None):
		s = Session()
		if s.has_changed and s.filename:
			ans = self.dlg_yesnocancel("There are unsaved changes. Would you like to save the current file before closing the application?")
			if ans == gtk.RESPONSE_YES:
				self.on_mnu_file_save_activate(widget, data)
			elif ans == gtk.RESPONSE_CANCEL:
				return True
		elif not s.filename and s.has_changed:
			ans = self.dlg_yesnocancel("There are unsaved changes. Would you like to save the current session?")
			if ans == gtk.RESPONSE_YES:
				self.on_mnu_file_save_as_activate(widget, data)
			elif ans == gtk.RESPONSE_CANCEL:
				return True

		gp = Gnuplot()
		gp.stop()
		self.builder.get_object("main").destroy()
		gtk.main_quit()

	def on_fc_data_file_file_set(self, widget, data = None):
		gp = Gnuplot()
		start = gp.get_last_event().strip() if gp.pwd() != None else os.curdir
		is_abs = self.child("chk_abs").get_active()
		path = '"%s"' % widget.get_filename() if is_abs else '"%s"' % os.path.relpath(widget.get_filename(), start)
		self.child("txt_function").set_text(path)

	def on_cbo_style_changed(self, widget, data = None):
		self.child("box_style").set_sensitive(bool(self.child("cbo_style").get_active_text().strip()))

	def on_chk_is_ls_toggled(self, widget, data = None):
		self.child("spb_ls").set_sensitive(self.child("chk_ls").get_active())
		self.child("box_ud").set_sensitive(not self.child("chk_ls").get_active())

	def on_chk_parametric_toggled(self, widget, data = None):
		gp = Gnuplot()
		gp.set_parametric() if widget.get_active() else	gp.unset("parametric")

	def on_mnu_file_new_activate(self, widget, data = None):
		s = Session()
		if s.has_changed and s.filename:
			ans = self.dlg_yesnocancel("There are unsaved changes. Would you like to save the current file before creating a new session?")
			if ans == gtk.RESPONSE_YES:
				self.on_mnu_file_save_activate(widget, data)
			elif ans == gtk.RESPONSE_CANCEL:
				return
		elif not s.filename and s.has_changed:
			ans = self.dlg_yesnocancel("There are unsaved changes. Would you like to save the current session?")
			if ans == gtk.RESPONSE_YES:
				self.on_mnu_file_save_as_activate(widget, data)
			elif ans == gtk.RESPONSE_CANCEL:
				return
		self.clear()
		s.new()
		self.update_window_title()

	def on_mnu_file_save_as_activate(self, widget, data = None):
		fc = gtk.FileChooserDialog("Save current workspace", self.window, \
		                           gtk.FILE_CHOOSER_ACTION_SAVE, \
		                           ("Cancel", gtk.RESPONSE_CANCEL,"OK", gtk.RESPONSE_OK))
		fc.set_do_overwrite_confirmation(True)
		resp = fc.run()
		if resp == gtk.RESPONSE_OK:
			s = Session()
			if s.save(fc.get_filename()):
				self.update_window_title()
		fc.destroy()

	def on_mnu_file_save_activate(self, widget, data = None):
		s = Session()

		if s.filename:
			if s.save():
				self.update_window_title()
		else:
			self.on_mnu_file_save_as_activate(widget, data)

	def on_mnu_file_open_activate(self, widget, data = None):
		fc = gtk.FileChooserDialog("Open an existing workspace", self.window, \
		                           gtk.FILE_CHOOSER_ACTION_OPEN, \
		                           ("Cancel",  gtk.RESPONSE_CANCEL, "OK", gtk.RESPONSE_OK))
		resp = fc.run()
		filename = fc.get_filename()
		fc.destroy()
		if resp == gtk.RESPONSE_OK:
			s = Session()
			if s.open(filename):
				self.clear()
				self.plot_from_hierarchy(s.ts_plot)
				
				gp = Gnuplot()
				gp.cmd(s.state)

			else:
				self.dlg_error("The specified file is not compatible with the current version of pyGtkPlot.")
				return

			try:
				gp = Gnuplot()
				match = re.match("(.*)[/](.*)", s.filename)
				gp.cd(match.group(1))
			except:
				pass
			s.has_changed = False
			self.update_window_title()
			#self.plot_list()

	def on_mnu_file_load_script_activate(self, widget, data = None):
		fc = gtk.FileChooserDialog("Open a gnuplot batch file", self.window, \
		                           gtk.FILE_CHOOSER_ACTION_OPEN, \
		                           ("Cancel",  gtk.RESPONSE_CANCEL, "OK", gtk.RESPONSE_OK))
		resp = fc.run()
		if resp == gtk.RESPONSE_OK:
			gp = Gnuplot()
			s = Session()
			gp.load(fc.get_filename())
			self.plot_from_last_plot_command()
			s.has_changed = True
			self.update_window_title()
			self.plot_list()
		fc.destroy()

	def on_mnu_file_cd_activate(self, widget, data = None):
		gp = Gnuplot()
		fc = gtk.FileChooserDialog("Change the current working directory", self.window, \
		                           gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER, \
		                           ("Cancel",  gtk.RESPONSE_CANCEL, "OK", gtk.RESPONSE_OK))
		if gp.pwd() != None: fc.set_filename(gp.get_last_event())
		resp = fc.run()
		if resp == gtk.RESPONSE_OK:
			filename = fc.get_filename()
			gp.cd(filename)
			self.status('Working directory changed to "%s"' % filename)
		fc.destroy()

	def on_mnu_file_export_activate(self, widget, data = None):
		exp = export.Export()
		resp = exp.run()
		if resp == gtk.RESPONSE_ACCEPT:
			#exp.export_current_view()
			filename, term, options = exp.get_output()
			gp = Gnuplot()
			gp.set_terminal("push")
			gp.set_terminal(term, options)
			gp.set_output(filename)
			self.plot_list()
			gp.set_output()
			gp.set_terminal("pop")
		exp.window.destroy()

	def on_mnu_file_render_activate(self, widget, data = None):
		ren = render.Render()
		resp = ren.run()
		if resp == gtk.RESPONSE_ACCEPT:
			ren.render_current_view()
		ren.window.destroy()

	def on_mnu_edit_select_all_activate(self, widget, data = None):
		self.child("tv_plot").get_selection().select_all()

	def on_mnu_edit_clear_activate(self, widget, data = None):
		self.clear()

	def on_mnu_edit_preferences_activate(self, widget, data = None):
		p = preferences.Preferences()
		p.window.set_transient_for(self.window)
		p.window.show_all()

	def on_mnu_view_appearances_activate(self, widget, data = None):
		app = appearances.Appearances()
		app.window.show_all()

	def on_mnu_view_ranges_activate(self, widget, data = None):
		ax = axes.Axes()
		ax.window.show_all()

	def on_mnu_mode_multiplot_button_press_event(self, widget, event, data = None):
		mp = multiplot.Multiplot()
		ans = mp.run()
		s = Session()
		gp = Gnuplot()
		if ans == gtk.RESPONSE_OK:
			s.is_multiplot = True
			s.multiplot_data = mp.get_layout()
			s.multiplot_total = len(s.multiplot_data)
			s.multiplot_count = mp.child("spb_start").get_value_as_int()
			self.child("mnu_mode_multiplot").set_active(True)
			gp.set_multiplot()
		else:
			s.is_multiplot = False
			s.multiplot_total = 0
			s.multiplot_count = 0
			self.child("mnu_mode_multiplot").set_active(False)
			gp.unset("multiplot")
			gp.set_size(1,1)
			gp.set_origin(0,0)
		mp.destroy()

	def on_mnu_tools_conformal_activate(self, widget, data = None):
		conf = conformal.Conformal(self)
		conf.show_all()

	def on_txt_raw_cmd_key_press_event(self, widget, event, data = None):
		s = Session()
		if (event.state, event.keyval) == (gtk.gdk.CONTROL_MASK, gtk.keysyms.Up):
			conf = config.Config()
			max_line = int(float(conf.max_history))
			if max_line:
				if len(s.history) > max_line > 0:
					s.history = s.history[-max_line:]
			else:
				s.history = []
			
			if len(s.history) == 0: return False
			
			self.show_history(self.populate_history(self.child("txt_raw_cmd").get_text()))
			return True

		elif event.keyval == gtk.keysyms.Escape:
			self.child("history").hide()

		return False

	def on_tv_history_row_activated(self, widget, path, column, data = None):
		ls, it = widget.get_selection().get_selected()

		if it:
			self.child("txt_raw_cmd").set_text(ls.get_value(it, 0))
			self.child("history").hide()

	def on_history_key_press_event(self, widget, event, data = None):
		if event.keyval == gtk.keysyms.Escape:
			self.child("history").hide()

		return False

	def on_txt_raw_cmd_changed(self, widget, data = None):
		try:
			if self.child("history").window.is_visible():
				conf = config.Config()
				self.populate_history(widget.get_text() if conf.filter_history == "True" else "")
		except:
			pass
